using System;
using UnityEngine;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;

namespace Vinput
{
    public class AutoThruster : MVRScript
    {
        public bool __DEBUG__ = false;

        // ============================================================================================ //
        // ======================================== PROPERTIES ======================================== //
        // ============================================================================================ //
        // UI variables
        private JSONStorableBool _activeJSON;
        private JSONStorableStringChooser _controlJSON;
        private JSONStorableFloat _distanceJSON;
        private JSONStorableFloat _distanceRandomJSON;
        private JSONStorableStringChooser _axisJSON;
        private JSONStorableBool _rotEnabledJSON;
        private JSONStorableFloat _rotTimeOffsetJSON;
        private JSONStorableFloat _rotStartXJSON;
        private JSONStorableFloat _rotStartYJSON;
        private JSONStorableFloat _rotStartZJSON;
        private JSONStorableFloat _rotEndXJSON;
        private JSONStorableFloat _rotEndYJSON;
        private JSONStorableFloat _rotEndZJSON;

        // Person atoms, controls, rigidbodies
        private FreeControllerV3 _control;

        // Internal
        private Cycler _cycler;
        private static readonly Dictionary<string, Vector3> _axisVectors = new Dictionary<string, Vector3>
        {
            {"X", Vector3.right},
            {"Y", Vector3.up},
            {"Z", Vector3.forward}
        };
        private float _currentDistance;
        private Vector3 _targetPos => _ctrlSnapshot.localPosition + (_ctrlSnapshot.localRotation * _axisVectors[_axisJSON.val]).normalized * _currentDistance / 10;
        private CtrlSnapshot _ctrlSnapshot;
        private bool _restoring = false;

        // ============================================================================================ //
        // ======================================== METHODS =========================================== //
        // ============================================================================================ //

        // BUILT-IN
        // Called by game engine on initialization
        public override void Init()
        {
            try
            {
                _cycler = new Cycler(this, OnBegin: ThrustBegin, OnUpdate: ThrustUpdate);
                SetupUI();
            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(AutoThruster)}.{nameof(Init)}: {e}");
            }
        }

        // Called by game engine before first update
        private void Start()
        {
            _UpdateState();
        }

        // Events/callbacks
        public void OnEnable() { _UpdateState(); }
        public void OnDisable() { _Deactivate(); }
        public void OnDestroy()
        {
            if (_cycler != null && _cycler.IsActive) _cycler.StopCycle();
            _RestoreCtrlSnapshot(_ctrlSnapshot);
        }

        // PLUGIN
        // Cycler dependencies
        private void ThrustBegin(Cycler _)
        {
            _currentDistance = UnityEngine.Random.Range(_distanceJSON.val * (1 - _distanceRandomJSON.val), _distanceJSON.val * (1 + _distanceRandomJSON.val));
        }
        private void ThrustUpdate(Cycler cycler)
        {
            if (__DEBUG__) SuperController.LogMessage($"Updating thrust: pct = {cycler.CurrentPct}, t = {cycler.CurrentT}");

            _control.transform.localPosition = Vector3.LerpUnclamped(_ctrlSnapshot.localPosition, _targetPos, cycler.CurrentT);

            if (_rotEnabledJSON.val)
            {
                _control.transform.localRotation = _ctrlSnapshot.localRotation * Quaternion.LerpUnclamped(
                    Quaternion.Euler(_rotStartXJSON.val, _rotStartYJSON.val, _rotStartZJSON.val),
                    Quaternion.Euler(_rotEndXJSON.val, _rotEndYJSON.val, _rotEndZJSON.val),
                    cycler.T((cycler.CurrentPct + _rotTimeOffsetJSON.val + 1) % 1)
                );
            }
        }

        // Activation/deactivation
        private void _UpdateState()
        {
            if (!_restoring)
            {
                if (_activeJSON.val && enabled)
                {
                    // Try to activate
                    _control = containingAtom.freeControllers.First(fc => fc.name == _controlJSON.val);
                    if (_ctrlSnapshot == null || _ctrlSnapshot.name != _control.name)
                    {
                        _RestoreCtrlSnapshot(_ctrlSnapshot);
                        _ctrlSnapshot = new CtrlSnapshot(_control);
                    }

                    _cycler.StartCycle();
                }
                else _Deactivate();
            }
        }

        private void _Deactivate()
        {
            if (!_cycler.IsActive) return;

            _cycler.StopCycle();
            _RestoreCtrlSnapshot(_control);
            _ctrlSnapshot = null;
        }


        private void OnControlChange(string _) { _UpdateState(); }
        private void OnActiveChange(bool _) { _UpdateState(); }
        private void OnEnableRotationsChange(bool val)
        {
            if (!val && _cycler.IsActive) _control.transform.localRotation = _ctrlSnapshot.localRotation;
        }

        // Save/restore
        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
            var json = base.GetJSON(includePhysical, includeAppearance, forceStore);

            var snapshot = Serializer.Serialize(_ctrlSnapshot);

            if (snapshot != null) json["CtrlSnapshot"] = snapshot;

            needsStore = true;
            return json;
        }

        public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true)
        {
            _restoring = true;
            base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms, setMissingToDefault);
            if (jc.HasKey("CtrlSnapshot")) _ctrlSnapshot = Serializer.DeserializeCtrlSnapshot(jc["CtrlSnapshot"] as JSONClass);

            _restoring = false;
        }
        private void _RestoreCtrlSnapshot(FreeControllerV3 control)
        {
            if (control == null) return;

            if (control.name == _ctrlSnapshot.name) _RestoreCtrlSnapshot(_ctrlSnapshot);
        }

        private void _RestoreCtrlSnapshot(CtrlSnapshot snapshot)
        {
            if (snapshot == null) return;

            FreeControllerV3 ctrl = containingAtom.freeControllers.First(fc => fc.name == snapshot.name);
            if (ctrl != null)
            {
                ctrl.transform.localPosition = snapshot.localPosition;
                ctrl.transform.localRotation = snapshot.localRotation;
            }
        }

        // UI
        private void SetupUI()
        {
            _activeJSON = UI_Utils.SetupToggle(this, "<color=#1010A0><size=32><b>Active</b></size></color>", false);
            _activeJSON.setCallbackFunction += OnActiveChange;

            // Control
            _controlJSON = new JSONStorableStringChooser("ControlName", Scene_Utils.ControlNames(containingAtom), Scene_Utils.ControlNames(containingAtom)[0], "Control");
            RegisterStringChooser(_controlJSON);
            _controlJSON.setCallbackFunction += OnControlChange;
            CreateScrollablePopup(_controlJSON).popup.onOpenPopupHandlers += () => { _controlJSON.choices = Scene_Utils.ControlNames(containingAtom); };

            // Settings - Distances
            CreateSpacer().height = 40f;
            _axisJSON = new JSONStorableStringChooser("ThrustAxis", _axisVectors.Keys.ToList(), "Z", "Thrust Axis");
            RegisterStringChooser(_axisJSON);
            CreateScrollablePopup(_axisJSON).popup.onOpenPopupHandlers += () => { _axisJSON.choices = _axisVectors.Keys.ToList(); };
            _distanceJSON = UI_Utils.SetupSlider(this, "Distance", 2.0f, -5.0f, 5.0f, false);
            _distanceRandomJSON = UI_Utils.SetupSlider(this, "Distance Random %", 0f, 0f, 0.5f, false);

            // Settings - Thrust
            CreateSpacer().height = 40f;
            _cycler.SetupUI(this);

            // Settings - Rotations
            CreateSpacer().height = 40f;
            _rotEnabledJSON = UI_Utils.SetupToggle(this, "Enable Rotations", true);
            _rotEnabledJSON.setCallbackFunction += OnEnableRotationsChange;
            _rotTimeOffsetJSON = UI_Utils.SetupSlider(this, "Rotation Time Offset %", 0f, -0.5f, 0.5f);
            _rotStartXJSON = UI_Utils.SetupSlider(this, "Rotation X Start", 0f, -90f, 90f);
            _rotEndXJSON = UI_Utils.SetupSlider(this, "Rotation X End", 0f, -90f, 90f);
            _rotStartYJSON = UI_Utils.SetupSlider(this, "Rotation Y Start", 0f, -90f, 90f);
            _rotEndYJSON = UI_Utils.SetupSlider(this, "Rotation Y End", 0f, -90f, 90f);
            _rotStartZJSON = UI_Utils.SetupSlider(this, "Rotation Z Start", 0f, -90f, 90f);
            _rotEndZJSON = UI_Utils.SetupSlider(this, "Rotation Z End", 0f, -90f, 90f);

            // Instructions
            string instructionText =
                "<color=#202020><size=36><b>Settings Information</b></size></color>\n\n" +
                "<b>Control</b>: Name of the control to thrust.\n\n" +

                "<color=#101060><b>Thrust Settings</b></color>\n" +
                "<b>Thrust Axis</b>: The axis of the control along which to thrust (X = Red, Y = Green, Z = Blue).\n" +
                "<b>Distance</b>: The distance to thrust. Set this to a negative value to thrust in the negative direction of the Thrust Axis.\n" +
                "<b>Distance Random %</b>: Proportion of Distance to randomly add/subtract from each cycle. A value of 0.0 means no random variation, a value of 0.5 means distances will randomly vary between 50% and 150% of Distance.\n\n" +

                "<color=#101060><b>Cycle Settings</b></color>\n" +
                "<b>Cycle Duration:</b> Duration of a complete <i>In</i> and <i>Out</i> motion.\n" +
                "<b>Cycle Duration % In:</b> Proportion of Cycle Duration taken by the <i>In</i> motion.\n" +
                " 0.5 = <i>In</i> and <i>Out</i> have equal duration.\n" +
                "Less than 0.5 = <i>In</i> motion is faster than <i>Out</i> motion.\n" +
                "More than 0.5 = <i>In</i> motion is slower than <i>Out</i> motion.\n" +
                "<b>Cycle Duration Random %:</b> Proportion of Cycle Duration to randomly add/subtract from each cycle. A value of 0.0 means no random variation, a value of 0.5 means cycles will randomly vary between 50% and 150% of Cycle Duration.\n" +
                "<b>Ease In:</b> Easing function for <i>In</i> motion. See https://easings.net for reference.\n" +
                "<b>Ease Out:</b> Easing function for <i>Out</i> motion.\n\n" +

                "<color=#101060><b>Rotation Settings</b></color>\n" +
                "<b>Rotation Time Offset %</b>: When set to 0, the control will reach the Start rotation values at the start of the thrust and the End values at the end of the thrust. Positive values 'shift' the rotations forward in time, negative values 'shift' them backwards\n" +
                "<b>Rotation X/Y/Z Start</b>: Relative rotations at the start of the thrust.\n" +
                "<b>Rotation X/Y/Z End</b>: Relative rotations at the end of the thrust.\n";
            CreateTextField(new JSONStorableString("text", instructionText), true).height = 1200;
        }
    }
}
